"""Test integrated plots."""
from pathlib import Path

import numpy as np

from taurenmd.plots import labeldots, plotparams


def test_plotparams_0():
    """Test plotparams defaults."""
    plotparams.plot(
        list(range(100)),
        list(range(100)),
        )

    assert Path('plot_param.pdf').exists()
    Path('plot_param.pdf').unlink()
    return


def test_plotparams_1():
    """Test plotparams swapped arguments."""
    plotparams.plot(
        list(range(100)),
        list(range(100)),
        labels='This labels',
        title='Title',
        xlabel='data',
        ylabel='observable',
        colors=['green'],
        xmax=50,
        xmin=20,
        ymax=50,
        ymin=20,
        grid=False,
        legend=False,
        filename='this_plot.png',
        )

    # py37 compatibility
    assert Path('this_plot.png').exists()
    Path('this_plot.png').unlink()
    return


def test_plotparams_2():
    """Test plotparams swapped arguments with multidimensional y-data."""
    plotparams.plot(
        list(range(50)),
        np.arange(100).reshape(2, 50),
        labels=['label 1', 'label 2'],
        title='Title',
        xlabel='data',
        ylabel='observable',
        colors=['green', 'blue'],
        xmax=50,
        xmin=20,
        ymax=50,
        ymin=20,
        grid=False,
        legend=False,
        vert_lines=(1, 10),
        filename='this_plot.png',
        )

    # py37 compatibility
    assert Path('this_plot.png').exists()
    Path('this_plot.png').unlink()
    return


def test_plotparams_3():
    """Test plotparams swapped arguments with multidimensional y-data."""
    plotparams.plot(
        list(range(50)),
        np.arange(100).reshape(2, 50),
        labels=['label 1', 'label 2'],
        title='Title',
        xlabel='data',
        xticks=list(range(50)),
        xticks_labels=list(map(str, range(50))),
        yticks=list(range(10)),
        yticks_labels=list(map(str, range(10))),
        hline=0.5,
        ylabel='observable',
        colors=['green', 'blue'],
        xmax=50,
        xmin=20,
        ymax=50,
        ymin=20,
        grid=False,
        legend=False,
        vert_lines=(1, 10),
        filename='this_plot.png',
        )

    # py37 compatibility
    assert Path('this_plot.png').exists()
    Path('this_plot.png').unlink()
    return


def test_labeldots_0():
    """Test label dots defaults."""
    labeldots.plot(
        list(range(50)),
        )

    # py37 compatibility
    assert Path('plot_param.pdf').exists()
    Path('plot_param.pdf').unlink()


def test_labeldots_1():
    """Test label dots defaults."""
    labeldots.plot(
        list(range(50)),
        x_labels=[f'{i}_' for i in range(50)],
        xlabel='x label',
        x_label_rot=45,
        labels=['series A'],
        title='title',
        legend=False,
        grid=False,
        filename='plot_test.png',
        )

    # py37 compatibility
    assert Path('plot_test.png').exists()
    Path('plot_test.png').unlink()


def test_labeldots_2():
    """Test label dots defaults."""
    labeldots.plot(
        np.arange(100).reshape(2, 50),
        x_labels=[f'{i}_' for i in range(50)],
        x_label_rot=45,
        colors=['k', 'w'],
        labels=['series A', 'series B'],
        title='title',
        legend=False,
        grid=False,
        filename='plot_test.png',
        )

    # py37 compatibility
    assert Path('plot_test.png').exists()
    Path('plot_test.png').unlink()



"""Plot a single parameter."""
import itertools

import numpy as np
from matplotlib import pyplot as plt

from taurenmd.libs import libutil


def plot(
        x_data,
        y_data,
        *,
        labels="No label provided",
        title=None,
        xlabel=None,
        ylabel=None,
        xlabel_fs=8,
        ylabel_fs=8,
        xticks=None,
        yticks=None,
        xticks_labels=None,
        yticks_labels=None,
        colors=('b', 'g', 'r', 'c', 'm', 'y', 'k'),
        alpha=0.7,
        xmax=None,
        xmin=None,
        ymax=None,
        ymin=None,
        hline=None,
        hline_color="black",
        hline_lw=1,
        xticks_fs=10,
        yticks_fs=10,
        grid=True,
        grid_color="lightgrey",
        grid_ls="-",
        grid_lw=1,
        grid_alpha=0.5,
        legend=True,
        legend_fs=8,
        legend_loc=0,
        vert_lines=None,
        figsize=(8, 5),
        filename='plot_param.pdf',
        dpi=150,
        ):
    """
    Plot a parameter.

    Bellow parameters concern data representation and are considered
    of highest importance because their incorrect use can mislead
    data analysis and consequent conclusions.

    Plot style parameters concernning only plot style, i.e., colors,
    shapes, fonts, etc... and which do not distort the actual data,
    are not listed in the paremeter list bellow. We hope these
    parameter names are self-explanatory and are listed in the function
    definition.

    Parameters
    ----------
    x_data : interable of numbers
        Container of the X axis data. Should be accepted
        by matplotlib.

    y_data : np.ndarray, shape=(M, len(x_data))
        Container of the Y axis data.
        Where M is the number of series.

    labels : str, optional
        The label to represent in plot legend.
        If a list of series is provided, a list of labels
        can be provided as well.
        Defauts to: "no labels provided".

    filename : str, optional
        The file name with which the plot figure will be saved
        in disk. Defaults to rmsd_individual_chains_one_subplot.pdf.
        You can change the file type by specifying its extention in
        the file name.

    fig_size : tuple of float or int
        The size ratio of the subplot in the figure.
    """
    y_data = np.array(y_data)

    if y_data.ndim == 1:
        y_data = y_data[np.newaxis, :]

    plot_labels = libutil.make_list(labels)
    plot_colors = libutil.make_list(colors)
    plot_colors = itertools.cycle(plot_colors)

    assert len(x_data) == y_data.shape[1], \
        '{} vs {}'.format(len(x_data), y_data.shape[1])
    assert y_data.shape[0] == len(plot_labels), (
        '{} vs {}'.format(y_data.shape[0], len(plot_labels)))

    fig, ax = plt.subplots(
        nrows=1,
        ncols=1,
        figsize=figsize,
        constrained_layout=False,
        )
    plt.tight_layout(rect=[0.05, 0.05, 0.995, 1])

    ax.set_title(
        title,
        x=0.5,
        y=1,
        va="bottom",
        ha="center",
        fontweight='bold',
        )

    for yy, ll in zip(y_data, plot_labels):
        ax.plot(
            x_data,
            yy,
            label=ll,
            color=next(plot_colors),
            alpha=alpha,
            zorder=2,
            )

    ax.set_xlabel(xlabel, weight='bold', fontsize=xlabel_fs)
    ax.set_ylabel(ylabel, weight='bold', fontsize=ylabel_fs)

    # setting axes scales

    xmin = xmin if xmin is not None else x_data[0]
    xmax = xmax if xmax is not None else x_data[-1]
    ymin = ymin if ymin is not None else np.array(y_data).min()
    ymax = ymax if ymax is not None else np.array(y_data).max()

    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)

    if xticks:
        ax.set_xticks(ticks=xticks)

    if xticks_labels:
        ax.set_xticklabels(xticks_labels)

    if yticks:
        ax.set_yticks(ticks=yticks)

    if yticks_labels:
        ax.set_yticklabels(yticks_labels)

    ax.tick_params(axis="x", labelsize=xticks_fs)
    ax.tick_params(axis="y", labelsize=yticks_fs)

    if grid:
        ax.grid(
            color=grid_color,
            linestyle=grid_ls,
            linewidth=grid_lw,
            alpha=grid_alpha,
            zorder=1,
            )

    if hline is not None:
        ax.axhline(hline, zorder=1, color=hline_color, lw=hline_lw)

    if isinstance(vert_lines, (list, tuple)):
        for line in vert_lines:
            ax.axvline(x=float(line), color='k')

    if legend:
        ax.legend(
            fontsize=legend_fs,
            loc=legend_loc,
            )
    plt.subplots_adjust(top=.9)
    fig.savefig(filename, dpi=dpi)

    plt.close("all")

    return



"""
# Calculate RMSFS of a selection along the trajectory slice.

## Algorithm

Calculates the RMSF values along a trajectory slice for different
selections. If multiple selections are given creates a series data
for that selection.

RMSF is calculated using [libcalc.mda_rmsf](https://taurenmd.readthedocs.io/en/latest/reference/libcalc.html#taurenmd.libs.libcalc.mda_rmsf).

If multiple selections are given, separate calculations are performed
in sequence. Result files (data tables and plots) are exported
separately for each selection. Selections can't be overlayed easily
in a single plot because they do not share the same labels.

## Examples

Calculate RMSF of the whole system:

    taurenmd rmsf top.pdb traj.dcd -e rmsf.csv

Calculates RMSFs for different selections:

    taurenmd rmsf top.pdb traj.dcd -g 'segid A' 'segid B' -e

``-x`` exports the data to a CSV file. You can also plot the data with
the ``-v`` option:

    [...] -x rmsf.csv -v title=my-plot-title xlabel=frames ylabel=RMSFs ...

where ``[...]`` is the previous command example.

you can also use ``tmdrmsf`` instead of ``taurenmd rmsf``.

## References

"""  # noqa: E501
import argparse
import functools
import os
from datetime import datetime

import numpy as np

from taurenmd import _BANNER, Path
from taurenmd import core as tcore
from taurenmd import log
from taurenmd.libs import libcalc, libcli, libmda, libutil
from taurenmd.libs.libutil import make_list
from taurenmd.logger import S, T
from taurenmd.plots import labeldots


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

__doc__ += (
    f'{tcore.ref_mda}'
    f'{tcore.ref_mda_selection}'
    )

_help = 'Calculate RMSFS for a selection and trajectory slice.'
_name = 'rmsf'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectories_arg(ap)
libcli.add_insort_arg(ap)
libcli.add_atom_selections_arg(ap)
libcli.add_inverted_array(ap)
libcli.add_slice_arg(ap)
libcli.add_data_export_arg(ap)
libcli.add_plot_arg(ap)


def _ap():
    return ap


def main(
        topology,
        trajectories,
        insort=False,
        start=None,
        stop=None,
        step=None,
        selections=('protein and name CA',),
        inverted_selections=None,
        export=False,
        plot=False,
        plotvars=None,
        **kwargs
        ):
    """Execute client main logic."""
    print(inverted_selections)
    log.info(T('starting RMSFs calculation'))

    topology = Path(topology)
    trajectories = [Path(t) for t in trajectories]

    u = libmda.load_universe(topology, *trajectories, insort=insort)

    frame_slice = libmda.get_frame_slices(u, start, stop, step)

    selections_list = make_list(selections)
    print(selections_list)
    if not selections_list:
        emsg = (
            'No selections defined. Check your `selections` argument input: '
            f'{selections}'
            )
        raise ValueError(emsg)

    log.info(T('calculating RMSFs'))

    labels = []
    rmsfs = []

    # Create a list of slice objects according to `inverted_selections`
    # If true, create [::-1] slice, and the data for that selection will be
    # inverted. This is useful for plotting DNA or other antiparallel chains.
    if inverted_selections:
        inv_sels = [
            slice(None, None, -1) if bool(int(inv)) else slice(None, None, None)
            for inv in inverted_selections
            ]
    else:
        inv_sels = [slice(None, None, None)] * len(selections_list)

    for i, sel in enumerate(selections_list):
        log.info(S('for sel: {}', sel))

        atom_group = u.select_atoms(sel)

        _ = libmda.draw_atom_label_from_atom_group(atom_group)[inv_sels[i]]
        labels.append(_)

        _ = libcalc.mda_rmsf(atom_group, frame_slice=frame_slice)[inv_sels[i]]
        rmsfs.append(_)

    if export:

        data_str = (
            "# Date: {1}{0}"
            "# Topology: {2}{0}"
            "# Trajectories {3}{0}"
            "# Atom,RMSF combinations{0}"
            "# {4}{0}"
            "{5}"
            ).format(
                os.linesep,
                datetime.now().strftime("%d/%m/%Y, %H:%M:%S"),
                Path(topology).resolve(),
                ','.join(f.resolve().str() for f in trajectories),
                ','.join(f'{sel},RMSF' for sel in selections_list),
                libutil.make_csv_lines_in_interleaved_manner(rmsfs, labels),
                )

        log.info(T('Saving data'))
        log.info(S('to: {}', export))
        Path(export).write_text(data_str)

    if plot:
        log.info(T("Plotting results:"))

        armsfs = np.array(rmsfs)
        ymax = np.max(armsfs)

        cli_defaults = {
            'ymax': ymax * 1.1 if ymax > 0 else ymax * 0.9,
            'filename': 'plot_rmsfs.pdf',
            'title': 'RMSFs',
            'xlabel': 'Atoms',
            'ylabel': r'RMSF ($\AA$)',
            'x_labels': [_l.split('.')[-1] for _l in labels[0]],
            'labels': selections_list,
            }

        cli_defaults.update(plotvars or dict())

        labeldots.plot(armsfs, **cli_defaults)

        log.info(S(f'saved plot: {cli_defaults["filename"]}'))

    return


maincli = functools.partial(libcli.maincli, ap, main)


if __name__ == '__main__':
    maincli()



"""Test libutil."""
import os
from math import isclose

import pytest

from taurenmd.libs import libutil


@pytest.mark.parametrize(
    's,expected',
    [
        ['00', (0.0, '')],
        ['0', (0.0, '')],
        ['-0', (0.0, '')],
        ['0.', (0.0, '')],
        ['.0ns', (0.0, 'ns')],
        ['12.3ps', (12.3, 'ps')],
        ['12.34minute', (12.34, 'minute')],
        ['-12.34', (-12.34, '')],
        ['10E40sd', (float('10E40'), 'sd')],
        ['10e40sd', (float('10e40'), 'sd')],
        ['1E4', (float('1E4'), '')],
        ['1E-4hours', (float('1E-4'), 'hours')],
        ['-10E40nanosecond', (float('-10E40'), 'nanosecond')],
        ['-10E-40', (float('-10E-40'), '')],
        ['-10e-40', (float('-10e-40'), '')],
        ['-.1E-4', (float('-.1E-4'), '')],
        ['10.2E30', (float('10.2E30'), '')],
        ['10.2e30', (float('10.2e30'), '')],
        ['.10E30', (float('.10E30'), '')],
        ['-10.2E30', (float('-10.2E30'), '')],
        ['-.10E30', (float('-.10E30'), '')],
        ]
    )
def test_string_to_timestep(s, expected):
    """Test correct conversion."""
    time_, unit_ = libutil.split_time_unit(s)
    assert isclose(time_, expected[0])
    assert unit_ == expected[1]


@pytest.mark.parametrize(
    's',
    [
        '12.34 # with comment',
        '12.34.12fdsfdsfds',
        '1E4.4',
        '.10E30E',
        '10.2.E30',
        '123#with wrong comment',
        'E10',
        'e10',
        '.E10',
        '-e',
        '-E10',
        '10-10E19',
        ]
    )
def test_string_to_timestep_wrong(s):
    """Test correct conversion."""
    with pytest.raises(ValueError):
        libutil.split_time_unit(s)


def test_interleaved_str_for_csv():
    """Test making interleaved CSV strings."""
    indata = [
        [1, 2, 3, 4],
        [5, 6, 7, 8],
        ]

    inlabels = [
        ['a', 'b', 'c', 'd'],
        ['e', 'f', 'g', 'h'],
        ]

    result = libutil.make_csv_lines_in_interleaved_manner(indata, inlabels)

    expected = (
        f'a,1,e,5{os.linesep}'
        f'b,2,f,6{os.linesep}'
        f'c,3,g,7{os.linesep}'
        'd,4,h,8'
        )

    assert result == expected


def test_interleaved_str_for_csv_2():
    """Test making interleaved CSV strings."""
    indata = [
        [1, 2, 3, 4],
        [5, 6, 7],
        ]

    inlabels = [
        ['a', 'b', 'c', 'd'],
        ['e', 'f', 'g'],
        ]

    result = libutil.make_csv_lines_in_interleaved_manner(indata, inlabels)

    expected = (
        f'a,1,e,5{os.linesep}'
        f'b,2,f,6{os.linesep}'
        f'c,3,g,7{os.linesep}'
        'd,4'
        )

    assert result == expected


@pytest.mark.parametrize(
    'indata,inlabels',
    [
        (
            [
                [1, 2, 3, 4],
                [5, 6, 7],
                ],
            [
                ['a', 'b', 'c', 'd'],
                ['e', 'f', 'g', 'e'],
                ]
            ),
        (
            [
                [1, 2, 3],
                [5, 6, 7, 8],
                ],
            [
                ['a', 'b', 'c', 'd'],
                ['e', 'f', 'g'],
                ]
            ),
        ]
    )
def test_interleaved_str_for_csv_3(indata, inlabels):
    """Test making interleaved CSV strings."""
    with pytest.raises(AssertionError):
        libutil.make_csv_lines_in_interleaved_manner(indata, inlabels)



"""
# Decompose Eurler angle rotations for a selection.

*Calculate the Roll, Pitch and Yaw angles of a plane along the trajectory.*

Read further on roll, pitch and yaw angles (Euler Angles) -
[wikipedia](https://en.wikipedia.org/wiki/Euler_angles).

Here we decompose these movements around the three different axis
centered at an origin.

## Calculation algorithm

Given a selection of three regions, ``selection A``, ``selection B``
and ``selection C``:

1. Center the system to the three selections center of geometry for
every frame, this is called the *origin*;

2. Calculate a plane given by the center of geometries of the three
selections, plane ABC;

3. Define the vector OA that goes from the origin to the center of
geometry of ``selection A``, this represents the Yaw axis;

4. Define the normal vector to the plane ABC (ABCn), this represents the
Roll axis;

5. Define the cross product betwee vectors OA and ABCn (AONn), this is the
Pitch axis;

6. Repeat this process for every frame.

Calculating the angles:

Angles represent the right hand rotation around an axis of the sistem in
a i-frame compared to the reference frame.

### Roll

The roll angle is given by the torsion angle between OA, origin, ABCn,
and OAi (in frame), displaced along ABCn.

### Pitch

The pitch angle is by the torsion angle between ABCn, origin, AONn, and
ABCni (in frame), displaced along AONn.

### Yaw

The yaw angle is given by the torsion angle between the AONn, origin,
OA, and AONni (in frame), displaced along OA.

## Examples

In the case of an homotrimer, define the axis and the origin on the trimers:

    taurenmd rotations -z 'segid A' 'segid B' 'segid C' -x rotations.csv

## References

"""  # noqa: E501
import argparse
import functools

import numpy as np

from taurenmd import _BANNER, Path
from taurenmd import core as tcore
from taurenmd import log
from taurenmd.libs import libcalc, libcli, libio, libmda
from taurenmd.logger import S, T
from taurenmd.plots import plotparams


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

__doc__ += (
    f'{tcore.ref_mda}'
    f'{tcore.ref_mda_selection}'
    )

_help = 'Calculates angular rotations across axes.'
_name = 'rotations'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectories_arg(ap)
libcli.add_insort_arg(ap)
libcli.add_plane_selection_arg(ap)
libcli.add_angle_unit_arg(ap)
libcli.add_reference_frame_arg(ap)
libcli.add_slice_arg(ap)
libcli.add_data_export_arg(ap)
libcli.add_plot_arg(ap)


def _ap():
    return ap


def main(
        topology,
        trajectories,
        plane_selection,
        insort=False,
        aunit='degrees',
        ref_frame=0,
        start=None,
        stop=None,
        step=None,
        export=False,
        plot=False,
        plotvars=None,
        **kwargs,
        ):
    """Execute main client logic."""
    log.info(T('starting'))

    topology = Path(topology)
    trajectories = list(map(Path, trajectories))
    u = libmda.load_universe(topology, *trajectories, insort=insort)

    frame_slice = libmda.get_frame_slices(u, start, stop, step)

    log.info(T('computing planes'))
    log.info(S('for reference frame: {}', ref_frame))

    ABC_selections = plane_selection  # variable renaming
    origin_selection = ' or '.join(ABC_selections)
    log.info(S('for selection: {!r}', origin_selection))

    # p stands for point
    # atomG stands for MDAnalysis AtomGroup
    pABC_atomG = u.select_atoms(origin_selection)
    pA_atomG = u.select_atoms(ABC_selections[0])
    pB_atomG = u.select_atoms(ABC_selections[1])
    pC_atomG = u.select_atoms(ABC_selections[2])

    u.trajectory[ref_frame]

    # defining the center of reference
    opABC_cog = pABC_atomG.center_of_geometry()
    log.info(S('Original pABC center of geometry: {}', opABC_cog))

    log.info(S(
        'transferring all coordinates of '
        'reference frame to the origin (0, 0, 0)'
        ))
    pABC_atomG.positions = pABC_atomG.positions - opABC_cog
    pABC_cog = pABC_atomG.center_of_geometry().copy()  # this should be zero
    log.info(S('New origin Center of Geometry'))
    log.info(S('pABC_cog: {}', pABC_cog))

    log.info(T('defining the reference axes'))
    # vector from the origin (0, 0, 0) to cog of each selection
    pA_cog = pA_atomG.center_of_geometry().copy()
    pB_cog = pB_atomG.center_of_geometry().copy()
    pC_cog = pC_atomG.center_of_geometry().copy()
    log.info(S('plane points definition:'))
    log.info(S('pA: {}', pA_cog))
    log.info(S('pB: {}', pB_cog))
    log.info(S('pC: {}', pC_cog))

    ref_plane_normal = libcalc.calc_plane_normal(pABC_cog, pA_cog, pB_cog)
    log.info(S('Normal vector to the reference plane: {}', ref_plane_normal))

    ref_plane_cross = np.cross(pA_cog, ref_plane_normal)

    log.info(S('Cross vector to pA and Normal vector: {}', ref_plane_cross))

    log.info(T('Calculating tilt angles'))
    total_frames = len(u.trajectory[frame_slice])
    log.info(S('for {} frames', total_frames))
    sizet = (total_frames, 3)
    roll_vectors = np.empty(sizet, dtype=np.float64)
    pitch_vectors = np.empty(sizet, dtype=np.float64)
    yaw_vectors = np.empty(sizet, dtype=np.float64)

    with libcli.ProgressBar(total_frames, suffix='frames') as pb:
        for i, _ts in enumerate(u.trajectory[frame_slice]):

            pABC_cog_ts = pABC_atomG.center_of_geometry().copy()
            pABC_atomG.positions = pABC_atomG.positions - pABC_cog_ts

            ts_positions = pABC_atomG.positions.copy()

            # get roll
            pABC_atomG.positions = ts_positions + ref_plane_normal
            roll_vectors[i, :] = pA_atomG.center_of_geometry().copy()

            # get pitch
            pABC_atomG.positions = ts_positions + ref_plane_cross
            pitch_vectors[i, :] = libcalc.calc_plane_normal(
                ref_plane_cross,
                pA_atomG.center_of_geometry(),
                pB_atomG.center_of_geometry(),
                )

            # get yaw
            pABC_atomG.positions = ts_positions + pA_cog
            pA_cog_yaw_ts = pA_atomG.center_of_geometry().copy()
            normalv = libcalc.calc_plane_normal(
                pA_cog,
                pA_cog_yaw_ts,
                pB_atomG.center_of_geometry(),
                )
            yaw_vectors[i, :] = np.cross(pA_cog_yaw_ts, normalv)
            pABC_atomG.positions = ts_positions

            pb.increment()

    log.info(S('done'))

    roll_torsion = libcalc.torsion_set(
        pA_cog,
        pABC_cog,
        ref_plane_normal,
        roll_vectors,
        )

    pitch_torsion = libcalc.torsion_set(
        ref_plane_normal,
        pABC_cog,
        ref_plane_cross,
        pitch_vectors,
        )

    yaw_torsion = libcalc.torsion_set(
        ref_plane_cross,
        pABC_cog,
        pA_cog,
        yaw_vectors,
        )

    if aunit == 'degrees':
        roll_torsion, pitch_torsion, yaw_torsion = \
            [np.degrees(a) for a in [roll_torsion, pitch_torsion, yaw_torsion]]

    if export:
        libio.export_data_to_file(
            list(range(len(u.trajectory))[frame_slice]),
            roll_torsion, pitch_torsion, yaw_torsion,
            fname=export,
            header=(
                '# Topology: {}\n'
                '# Trajectories: {}\n'
                '# Plane Selection: {}\n'
                '# frame,roll,pitch,yaw [in {}]\n'
                ).format(
                    topology.resolve().str(),
                    ', '.join(t.resolve().str() for t in trajectories),
                    origin_selection,
                    aunit,
                    ),
            )

    if plot:
        log.info(T("Plotting results:"))

        plotvars = plotvars or dict()

        xdata_in_time = plotvars.pop('xdata_in_time', 'ns')
        frame_list = libmda.get_frame_list_from_slice(u, frame_slice)
        xdata, xlabel = libmda.create_x_data(u, xdata_in_time, frame_list)

        subtitle = ' · '.join(plane_selection)
        ymax = max(map(np.max, [roll_torsion, pitch_torsion, yaw_torsion]))
        ymin = min(map(np.min, [roll_torsion, pitch_torsion, yaw_torsion]))

        cli_defaults = {
            'dpi': 600,
            'filename': 'plot_rotations.png',
            'labels': ['roll', 'pitch', 'yaw'],
            'legend': True,
            'title': f'Roll, pitch, and yaw angles\n{subtitle}',
            'xlabel': xlabel,
            'ylabel': aunit,
            'ymax': ymax * 1.1 if ymax > 0 else ymax * 0.9,
            'ymin': ymin * 1.1 if ymin < 0 else ymin * 0.9,
            }

        cli_defaults.update(plotvars or dict())

        plotparams.plot(
            xdata,
            np.vstack((roll_torsion, pitch_torsion, yaw_torsion)),
            **cli_defaults,
            )

        log.info(S(f'saved plot: {cli_defaults["filename"]}'))

    return


maincli = functools.partial(libcli.maincli, ap, main)


if __name__ == '__main__':
    maincli()



"""
Functions that wrap around `MDAnalysis library`_.

Functions contained in this module operate with MDAnalysis (MDA)
functionalities, either by using MDA to access Molecular Dynamics
data or by receiving MDA data structures and parsing them in some way.

When using functions contained in this library you should `cite both`_
taurenmd and MDAnalysis.

.. _MDAnalysis library: https://www.mdanalysis.org
.. _cite both: https://taurenmd.readthedocs.io/en/latest/citing.html
"""
import MDAnalysis as mda
from MDAnalysis.analysis import align as mdaalign

from taurenmd import Path
from taurenmd import core as tcore
from taurenmd import log
from taurenmd.libs import libio, libopenmm, libutil
from taurenmd.logger import S, T


@tcore.add_reference(tcore.ref_mda)
def load_universe(
        topology,
        *trajectories,
        insort=False,
        **universe_args):
    """
    Load MDAnalysis universe.

    Accepts MDAnalysis compatible `topology formats`_ and
    `trajectory formats`_. Read further on the `MDAnalysis Universe`_.

    .. _topology formats: https://www.mdanalysis.org/docs/documentation_pages/topology/init.html
    .. _trajectory formats: https://www.mdanalysis.org/docs/documentation_pages/coordinates/init.html#supported-coordinate-formats
    .. _MDAnalysis Universe: https://www.mdanalysis.org/docs/documentation_pages/core/universe.html?highlight=universe#core-object-universe-mdanalysis-core-universe

    Examples
    --------

        >>> libmda.load_universe('topology.pdb', 'trajectory.dcd')

        >>> libmda.load_universe(
                'topology.pdb',
                'traj_part_1.xtc',
                'traj_part_2.xtc',
                Path('my', 'md', 'folder', 'traj_part_3.xtc'),
                )

    Parameters
    ----------
    topology : str or Path object
        Path to topology file.

    trajectories* : str of Path objects
        Paths to trajectory file(s). Trajectory files will be used
        sequentially to create the Universe.

    insort : bool
        Whether to sort trajectory files by suffix number.
        See :func:`libio.sort_numbered_input`.

    universe_args : any
        Other arguments to be passed to `MDAnalysis Universe`.

    Return
    ------
    MDAnalysis Universe
    """  # noqa: E501 D412
    if insort:
        trajectories = libio.sort_numbered_input(*trajectories)

    libio.report_input(topology, trajectories)

    topo_path = Path(topology).str()
    traj_path = [Path(i).str() for i in trajectories],

    try:
        universe = mda.Universe(topo_path, traj_path, **universe_args)

    except ValueError:
        pdbx = libopenmm.attempt_to_load_top_from_simtk(topo_path)
        universe = mda.Universe(pdbx, traj_path, **universe_args)

    report(universe)
    return universe


@tcore.add_reference(tcore.ref_mda)
def report(universe):
    """
    Report information about the Universe.

    Example
    -------

        >>> u = libmda.load_universe('topology.pdb', 'trajectory.xtc')
        >>> libmda.report(u)

    Parameters
    ----------
    universe : MDAnalysis Universe
        `MDAnalysis universe <https://www.mdanalysis.org/docs/documentation_pages/core/universe.html?highlight=universe#core-object-universe-mdanalysis-core-universe>`_.

    Returns
    -------
    None

    """  # noqa: E501
    segids = sorted(list(set(universe.atoms.segids)))
    info = {}
    for segid in segids:
        a = universe.select_atoms(f'segid {segid}')
        info[segid] = sorted(list(set(a.atoms.resids)))

    info_ = []
    for k, v in info.items():
        info_.append(str(S(
            f'segid {k} with {len(v)} residues '
            f'from {v[0]} to {v[-1]}'
            )))

    log.info(T('Reporting on universe'))
    log.info(S('number of frames: {}', len(universe.trajectory)))
    dt = get_timestep(universe, i1=-1)
    log.info(S('duration: {:.2f} ns', dt / 1000))

    dt = get_timestep(universe)
    log.info(S('timestep per frame: {:.2f} ns', dt / 1000))
    log.info(S('number of atoms: {}', len(universe.atoms)))
    log.info(S('components:\n{}', '\n'.join(info_)))


@tcore.add_reference(tcore.ref_mda)
def mdaalignto(universe, reference, selection='all'):
    """
    Align universe to reference.

    Uses `MDAnalysis.analysis.align.alignto <https://www.mdanalysis.org/docs/documentation_pages/analysis/align.html?highlight=alignto#MDAnalysis.analysis.align.alignto>`_.

    Parameters
    ----------
    universe, reference, selection
        Same as in ``MDAnalysis.analysis.align.alignto`` function.

    Raises
    ------
    ZeroDivisionError
        If selection gives empty selection.
    """  # noqa: E501
    try:
        mdaalign.alignto(universe, reference, select=selection)
    except (ValueError, ZeroDivisionError) as err:
        log.debug(err, exc_info=True)
        errmsg = (
            f'Could not perform alignment due to {err}, '
            'most likely the alignment selection does not match '
            'any possible selection in the system. You selection : '
            f'\'{selection}\'.'
            )
        log.info(errmsg)
        raise err


@tcore.add_reference(tcore.ref_mda)
def draw_atom_label_from_atom_group(atom_group):
    """
    Translate MDAnalysis Atom Group to list of strings for each atom.

    Strings represent each atom by SEGID.RESNUM|RESNAME.NAME,
    for example carbon alpha of Cys 18 of chain A would:

        >>> A.18Cys.CA

    This function is used by taurenmd for data representation
    purporses.

    Parameters
    ----------
    atom_group : Atom Group obj
       `MDAnalysis Atom group <https://www.mdanalysis.org/docs/documentation_pages/core/groups.html?highlight=atom%20group#MDAnalysis.core.groups.AtomGroup>`_.

    Returns
    -------
    list of strings
        Containing the atom string representation for each atom in
        the Atom Group.
    """  # noqa: E501
    labels = []
    for atom in atom_group:
        s = '{}.{}{}.{}'.format(
            atom.segment.segid,
            atom.residue.resnum,
            atom.resname.title(),
            atom.name,
            )
        labels.append(s)

    return labels


@tcore.add_reference(tcore.ref_mda)
def get_timestep(u, i0=0, i1=1):
    """
    Get time step between two trajectory frames in picoseconds.

    Parameters
    ----------
    it0, it1 : int
        The initial and end frame, respectively.
    """
    return u.trajectory[i1].time - u.trajectory[i0].time  # in picoseconds


@tcore.add_reference(tcore.ref_mda)
def convert_time_to_frame(x, dt, base_unit='ps'):
    """
    Convert a string `x` into a frame number based on given `dt`.

    If `x` does not contain any units its assumed to be a frame number
    already.

    Original function taken from `MDAnalysis.mdacli` project, from
    commit: https://github.com/MDAnalysis/mdacli/blob/15f6981df7b14ef5d52d64b56953d276291068ab/src/mdacli/utils.py#L25-L66
    The original function was modified internally without modifying its
    API.

    See also: https://github.com/MDAnalysis/mdacli/pull/81

    Note that in case of decimal values, for example, 10.3ps, the
    integer division between the value and `dt` remains.

    Parameters
    ----------
    x : str, int or float
        the input string

    dt : float
        the time step in ps

    Returns
    -------
    int
        frame number

    Raises
    ------
    ValueError
        The input does not contain any units but is not an integer.
    """  # noqa: E501
    x = str(x)

    # regex to split value and units while handling scientific input
    val, unit = libutil.split_time_unit(x)
    print(val, unit, dt)

    if unit != "":
        converted = mda.units.convert(val, unit, base_unit)
        return round(converted / dt, 0)
    else:
        return int(val)


def convert_time_or_frame_to_frame(s, u):
    """
    Convert a time or frame string to frame.

    Parameters
    ----------
    s : str or int or float
        A string defining the time ('12ns') or frame.

    u : mda.Universe

    Returns
    -------
    int
        see :func:`convert_time_to_frame`
    """
    dt = get_timestep(u)
    return convert_time_to_frame(s, dt)


def get_frame_list_from_slice(u, frame_slice):
    """
    Create a frame number list from a slice for a Universe.

    Parameters
    ----------
    u : mda.Universe
        The MDAnalysis universe.

    frame_slice : slice object

    Returns
    -------
    list of ints
        A list with the number of frames corresponding to that Universe
        and slice.
    """
    return list(range(len(u.trajectory))[frame_slice])


def get_frame_slices(u, start=None, stop=None, step=None):
    """Make a frame slice for a universe given start, stop, and step.

    Parameters
    ----------
    start, stop, step: None or int or str.
        If string, accepts :func:`libutil.split_time_unit`.

    Returns
    -------
    slice object
    """
    log.info(T('Creating trajectory frame slicing from:'))
    log.info(S(f'start: {start}'))
    log.info(S(f'stop : {stop}'))
    log.info(S(f'step : {step}'))

    frame_slice = libio.frame_slice(
        start=None if start is None else convert_time_or_frame_to_frame(str(start), u),  # noqa: E501
        stop=None if stop is None else convert_time_or_frame_to_frame(str(stop), u),  # noqa: E501
        step=None if step is None else convert_time_or_frame_to_frame(str(step), u),  # noqa: E501
        )

    log.info(S('created slice: {}', frame_slice))

    assert isinstance(frame_slice, slice)
    return frame_slice


def create_x_data(u, xdata_in_time, frame_list):
    """
    Create X data for plotting.

    Create a frame list given a time or else return the frame_list given.

    Parameters
    ----------
    u : mda.Universe

    xdata_in_time : bool
        Whether to select frames based on a time duration.

    frame_list : list of int
        A list of integers referring to the number of frames in the
        trajectory.
    """
    if xdata_in_time:
        xdata = [
            mda.units.convert(get_timestep(u, i1=i), 'ps', xdata_in_time)
            for i in frame_list
            ]
        xlabel = f'Time ({xdata_in_time})'
    else:
        xdata = frame_list
        xlabel = 'Frames'

    return xdata, xlabel



"""Test libmda."""
import copy
from math import isclose

import MDAnalysis as mda
import numpy as np
import pytest

from taurenmd.libs import libmda as la

from . import toptest, toptest_cif, trajtest


@pytest.fixture
def universe():
    """Return a Universe."""
    return la.load_universe(toptest, trajtest)


def test_load_universe_1(universe):
    """Test load MDA Universe."""
    assert isinstance(universe, mda.Universe)
    assert len(universe.trajectory) == 10


def test_load_universe_1_cif():
    """Test load MDA Universe."""
    universe = la.load_universe(toptest_cif, trajtest)
    assert isinstance(universe, mda.Universe)
    assert len(universe.trajectory) == 10


def test_load_universe_2():
    """Test load MDA Universe multiple trajs."""
    universe = la.load_universe(
        toptest,
        trajtest,
        trajtest,
        trajtest,
        )
    assert isinstance(universe, mda.Universe)
    assert len(universe.trajectory) == 30


def test_load_universe_2_insort():
    """Test load MDA Universe multiple trajs."""
    universe = la.load_universe(
        toptest,
        trajtest,
        trajtest,
        trajtest,
        insort=True,
        )
    assert isinstance(universe, mda.Universe)
    assert len(universe.trajectory) == 30


def test_load_universe_3():
    """Test load MDA Universe multiple trajs."""
    mapp = map(str, [trajtest] * 3)
    universe = la.load_universe(toptest, *mapp)
    assert isinstance(universe, mda.Universe)
    assert len(universe.trajectory) == 30


def test_report_universe(universe):
    """
    Test universe report.

    Only interface, does not test return value.
    """
    la.report(universe)


def test_mdaalignto(universe):
    """Test MDA alignto."""
    reference = mda.Universe(toptest.str())
    ugroup = universe.select_atoms('all')

    for _ts in universe.trajectory[1:]:
        prev = copy.copy(ugroup.positions)
        assert np.all(np.equal(prev, ugroup.positions))
        la.mdaalignto(universe, reference, 'name CA')
        assert np.all(np.not_equal(prev, ugroup.positions))


def test_mdaalignto_exit(universe):
    """Test exit on ZeroDivision."""
    reference = mda.Universe(toptest.str())
    with pytest.raises((ValueError, ZeroDivisionError)):
        la.mdaalignto(universe, reference, 'name ZZZZZ')


def test_draw_atom_label(universe):
    """Test label maker."""
    atom_group = universe.select_atoms('resnum 10-12 and name CA')
    labels = la.draw_atom_label_from_atom_group(atom_group)
    assert isinstance(labels, list)
    assert labels == ['A.10Ser.CA', 'A.11Ile.CA', 'A.12Leu.CA']


@pytest.mark.parametrize(
    'x, frame',
    (
        ('1', 1),
        ('-1', -1),
        ('1ns', 1000),
        ('1.5ns', 1500),
        ('12e3', 12e3),
        ('12e3ps', 12e3),
        ),
    )
def test_convert_str_time(x, frame):
    """
    Test convert string to time.

    Tests taken from: https://github.com/MDAnalysis/mdacli/blob/15f6981df7b14ef5d52d64b56953d276291068ab/tests/test_utils.py#L22-L41
    """  # noqa: E501
    assert frame == la.convert_time_to_frame(x, dt=1)


@pytest.mark.parametrize(
    'time_,dt,expected',
    [
        ('10ps', 10, 1),
        ('0.05ns', 50, 1),
        ]
    )
def test_convert_str_time_dt(time_, dt, expected):
    """
    Test convert string to time in ps.

    Tests taken from: https://github.com/MDAnalysis/mdacli/blob/15f6981df7b14ef5d52d64b56953d276291068ab/tests/test_utils.py#L22-L41
    """  # noqa: E501
    assert expected == la.convert_time_to_frame(time_, dt=dt)


@pytest.mark.parametrize(
    'value',
    [
        'nothing',
        ''
        ],
    )
def test_convert_str_time_raise(value):
    """
    Test convert string to time ValueError.

    Tests taken from: https://github.com/MDAnalysis/mdacli/blob/15f6981df7b14ef5d52d64b56953d276291068ab/tests/test_utils.py#L22-L41
    """  # noqa: E501
    with pytest.raises(ValueError):
        la.convert_time_to_frame(value, dt=1)


@pytest.mark.parametrize(
    'timestep,expected',
    [
        ('ns', [0.001, 0.002, 0.003, 0.004, 0.005]),
        ('ps', [1, 2, 3, 4, 5]),
        ]
    )
def test_create_x_data(universe, timestep, expected):
    """Test create x_data."""
    xdata, xlabel = la.create_x_data(universe, timestep, list(range(1, 6)))
    assert xlabel == f'Time ({timestep})'
    assert all(isclose(i, j, rel_tol=0.00001) for i, j in zip(xdata, expected))


def test_create_x_data_w_frames(universe):
    """Test create x_data."""
    xdata, xlabel = la.create_x_data(universe, False, list(range(10)))
    assert xlabel == 'Frames'
    expected = list(range(10))
    assert all(isclose(i, j, rel_tol=0.001) for i, j in zip(xdata, expected))


@pytest.mark.parametrize(
    'f1, f2, expected',
    [
        (0, -1, 9),
        (0, 1, 1),
        (0, 0, 0),
        ]
    )
def test_get_time(universe, f1, f2, expected):
    """Test get time from Univerise."""
    result = la.get_timestep(universe, i0=f1, i1=f2)
    assert isclose(result, expected, rel_tol=0.001)


@pytest.mark.parametrize(
    'slc,expected',
    [
        (slice(3, None), [3, 4, 5, 6, 7, 8, 9]),
        (slice(None, None), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
        (slice(None, 3), [0, 1, 2]),
        (slice(3, None, 3), [3, 6, 9]),
        ]
    )
def test_get_frame_list_from_slice(universe, slc, expected):
    """Test get a frame list from slice."""
    result = la.get_frame_list_from_slice(universe, slc)
    assert result == expected


@pytest.mark.parametrize(
    's,expected',
    [
        ('0ps', 0),
        ('1ps', 1),
        ('2ps', 2),
        ('3ps', 3),
        ('4ps', 4),
        ('5ps', 5),
        ('6ps', 6),
        ('7ps', 7),
        ('8ps', 8),
        ('9ps', 9),
        ('10ps', 10),
        ('0.005ns', 5),
        ('4', 4),
        (3, 3),
        (0, 0),
        ]
    )
def test_get_time_or_frame_to_frame(universe, s, expected):
    """Test get frame."""
    result = la.convert_time_or_frame_to_frame(s, universe)
    assert result == expected


@pytest.mark.parametrize(
    'start,stop,step,expected',
    [
        (None, None, 1, slice(None, None, 1)),
        (2, 5, 1, slice(2, 5, 1)),
        (8, 15, 1, slice(8, 15, 1)),
        ('2ps', None, None, slice(2, None, None)),
        ('0.004ns', '8ps', None, slice(4, 8, None)),
        ]
    )
def test_get_frame_slices(universe, start, stop, step, expected):
    """Test get frame slices."""
    result = la.get_frame_slices(
        universe,
        start=start,
        stop=stop,
        step=step,
        )

    assert result == expected



"""General utilities."""
import os
import re
from itertools import zip_longest


def make_list(*items):
    """Make a list out of objects."""
    new = []

    for item in items:
        if isinstance(item, str):
            new.append(item)
        else:
            new.extend(item)
    return new


def split_time_unit(s):
    """
    Split time and units.

    Follows the regex: https://regex101.com/r/LZAbil/2

    Returns
    -------
    tuple (float, str)
        Value as float, units as str.

    Raises
    ------
    IndexError
        Tuple could not be found. This happens when a number is not
        present in the start of the string.
    """
    type_regex = re.compile(r'^(\-?\d+\.?\d*|\-?\.\d+|\-?\.?\d+[eE]\-?\d+|-?\d+\.?\d*[eE]\d+)($|[a-z]*$)')  # noqa: E501
    try:
        value, unit = type_regex.findall(s)[0]
    except IndexError as err:
        raise ValueError(f'Time string not appropriate: {s!r}') from err
    return float(value), unit


def make_csv_lines_in_interleaved_manner(data, labels):
    """
    Create CSV lines from a list of lists in interleaved manner.

    Labels and data do not need to be of the same size. Resulting
    sizes will be completed with empty strings for those series with
    less values.

    Parameters
    ----------
    data : list of lists of numbers
        The data series.

    labels : list of lists of str
        The labels refering to the data series.

    Returns
    -------
    str
        A string merging the lines.
    """
    # require
    assert all(len(i) == len(j) for i, j in zip(data, labels)), \
        'data and labels size do not match'

    dataz = zip_longest(*data, fillvalue='')
    labelz = zip_longest(*labels, fillvalue='')

    lines = []
    for di, li in zip(dataz, labelz):
        _lines = (f'{_l},{_d}' for _d, _l in zip(di, li) if _d)
        lines.append(','.join(_lines))
    return os.linesep.join(lines)



"""Plot a single parameter."""
import itertools

import numpy as np
from matplotlib import pyplot as plt

from taurenmd.libs import libutil


def plot(
        y_data,
        x_labels=None,
        x_label_rot=90,
        labels=None,
        ymax=None,
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        legend_fs=6,
        legend_loc=4,
        colors=('b', 'g', 'r', 'c', 'm', 'y', 'k'),
        alpha=0.7,
        grid=True,
        grid_color="lightgrey",
        grid_ls="-",
        grid_lw=1,
        grid_alpha=0.5,
        figsize=(10, 6),
        filename='plot_param.pdf',
        dpi=300,
        ):
    """Plot label dots template."""
    y_data = np.array(y_data)
    if y_data.ndim == 1:
        y_data = y_data[np.newaxis, :]

    plot_colors = itertools.cycle(libutil.make_list(colors))

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)

    if xlabel:
        plt.tight_layout(rect=[0.05, 0.15, 0.995, 0.985])
    else:
        plt.tight_layout(rect=[0.05, 0.10, 0.995, 0.985])

    fig.suptitle(
        title,
        x=0.5,
        y=0.990,
        va="top",
        ha="center",
        )

    labels = labels or [None] * len(y_data)
    for i, yy in enumerate(y_data):

        ax.scatter(
            range(len(yy)),
            yy,
            label=labels[i],
            color=next(plot_colors),
            alpha=alpha,
            zorder=10)

    ax.set_xlabel(xlabel, weight='bold')
    ax.set_ylabel(ylabel, weight='bold')

    ax.set_ylim((0, ymax or np.max(y_data)))

    ax.set_xticks(list(range(y_data.shape[1])))

    if x_labels:
        ax.set_xticklabels(x_labels, rotation=x_label_rot)

    if grid:
        ax.grid(
            which='major',
            color=grid_color,
            linestyle=grid_ls,
            linewidth=grid_lw,
            alpha=grid_alpha,
            zorder=1,
            )

    if legend:
        ax.legend(
            fontsize=legend_fs,
            loc=legend_loc,
            )

    fig.tight_layout()
    fig.savefig(filename, dpi=dpi)
